/* Standard-Definitions */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "matrix.h"
#include "vectors.h"
#include "apply_h.h"

/* The parameters */

#define DELTA 2.0L
#define JVAL  1.0L

#define LINSIZE 6
#define NPART   6

/* A special number (should be negative, otherwise arbitrary) */

#define MAGIC -12345

/* The maximal spatial dimension */

#define MAXDIM 100

/* The spatial dimension */

static int ddim;

/* The maximal linear extent */

#define MAXD 32

/* The linear size */

static int lin_size=LINSIZE;

/* The total number of sites */

static int sites;

/* The total number of words to describe the state of the system */

static int words;

/* The number of particles (corresponding to total S^z) */

static int n_part=NPART;

/* The number of states in the total-S^z vector space */

static long nstates;

/* Dimension of the momentum eigenspace */

static long pstates;

/* Mapping tables */

static long *charge_momentum;
	/* Which S^z eigenvector belongs to which momentum eigenvector ? */

static long *momentum_charge;
	/* Representative S^z eigenvector for the momentum eigenvector */

static int *momentum_prefac;
	/* The number of states associated to the momentum eigenvector */

/* Table of binomial coefficients */

static long *bin_coeff[1000];

/* ****** Some math routines ****** */

#define over(a,b) bin_coeff[a][b]
/*
#define over(a,b) binomial(a,b)
*/

/* Den groessten gemeinsamen Teiler zweier Zahlen feststellen
   ueber Euklidischen Algorithmus
   Zehnerpotenzen schneller als Primfaktorzerlegung */

static long get_gcd(long a, long b)
 {
   register int i;
   long remainder;

   if(a == 0)           /* Sonderbehandlung fuer Nullen */
     return(1);
   if(b == 0)
     return(1);

   if(a < 0)            /* Betrag bilden */
     a=-a;
   if(b < 0)
     b=-b;

   if(a == 1)           /* Sonderbehandlung fuer Einsen */
     return(1);
   if(b == 1)
     return(1);

   remainder = a % b;   /* Das ist der Euklische Algorithmus */
   while(remainder)
    {
      a=b;
      b=remainder;
      remainder = a % b;
    }           /* Nun steht der GGT in b */

   return(b);
 }

/* Der Binomialkoeffizient */

static long binomial(long n, long m)
 {
   register long i;
   register long gcd;
   register long res1, res2;

   if(m > n-m)                  /* Symmetrie ausnutzen */
     m = n-m;

   if((n==36) && (m==10))
     return(254186856L);


   if(m < 0)
     return(0);

   res1 = 1;
   res2 = 1;
   for(i=0; i<m; i++)
    {
      res1 *= (n-i);           /* Multiplizieren */
      res2 *= (i+1);
      gcd = get_gcd(res1, res2);      /* Und kuerzen */
      res1 /= gcd;
      res2 /= gcd;
    }
   if(res2 != 1)
     fprintf(stderr, "*** Error: Binomial coefficient (%ld,%ld) not an integer !\n", n, m);
   return(res1);
 }

/* Allocate memory and produce error message if not available */

static void *m_alloc(long size, char *str)
 {
  void *ptr;

  ptr = (void *)malloc(size);

  if(! ptr)
   {
    fprintf(stderr, "### Error: Out of memory in %s()\n", str);
    exit(-13);
   }
  return(ptr);
 }

/* Initialize binomial coefficients */

static void init_binomials()
 {
  register long n, m;

  for(n=0; n<=sites; n++)
   {
    bin_coeff[n] = (long *)m_alloc((((n+1)<n_part)?(n+2):(n_part+1))*sizeof(long), "init_binomials");
    for(m=0; m<=n+1; m++)
     if(m <= n_part)
       bin_coeff[n][m] = binomial(n,m);
   }
 }

/* Free binomial coefficients */

static void free_binomials()
 {
  register long n;

  for(n=0; n<=sites; n++)
    free(bin_coeff[n]);
 }

/* Generate states in S^z-symmetry subspace */

static void gen_states()
 {
  nstates = over(sites, n_part);
 }

/* Free states in S^z-symmetry subspace */

static void free_states()
 {
 }

/* Convert index to bit representation of state */

static unsigned long *index_to_state(long index)
 {
  unsigned long *state;
  unsigned long mask;
  int np, hh;
  int count;
  register int i, j, k;

  state = (unsigned long *) m_alloc(sizeof(unsigned long) * words, "index_to_state");
  for(j=0; j<words; j++)
    state[j] = 0;

  count = 0;
  np = n_part;
  for(j=0; j<words; j++)
   {
    mask = 1<<(lin_size-1);
    for(i=0; i<lin_size; i++)
     {
      hh = over(sites-count-1, np);
#ifdef DEBUG
      printf("(%d,%d) = %d\n", sites-count-1, np, hh);
#endif
      if(index >= hh)
       {
        index -= hh;
        np--;
        state[j] |= mask;
       }
      count++;
      mask = mask>>1;
     }
   }

  if(index || (np < 0))
   fprintf(stderr, "*** Error: failed to translate index to state\n");

#ifdef DEBUG
  fprintf(stderr, "Result of index_to_state:\n");
  for(k=0; k<words; k++)
     fprintf(stderr, "\t%lx\n", state[k]);
#endif

  return(state);
 }

/* Convert bit representation of state to index of matrix */

static long state_to_index(unsigned long *state)
 {
  long index;
  unsigned long mask;
  int np;
  int count;
  register int i, j, k;

  count = index = 0;
  np = n_part;
  for(j=0; j<words; j++)
   {
    mask = 1<<(lin_size-1);
    for(i=0; i<lin_size; i++)
     {
      if(state[j] & mask)
       {
        index += over(sites-count-1, np);
        np--;
        if(np < 0)
         {
          fprintf(stderr, "### Error: Illegal state encountered:\n");
          for(k=0; k<words; k++)
             fprintf(stderr, "\t%lx\n", state[k]);
         }
       }
      count++;
      mask = mask>>1;
     }
   }

#ifdef DEBUG
  fprintf(stderr, "State:\n");
  for(k=0; k<words; k++)
     fprintf(stderr, "\t%lx\n", state[k]);
  fprintf(stderr, "Index:%ld\n", index);
#endif

  return(index);
 }

/* Translate a state bitwise left */

static void translate_left(unsigned long *state)
 {
  register unsigned long mask;
  register int i;

  mask = ((unsigned long) (1<<lin_size)) - 1;
  if(! mask)
    mask = -1;
  for(i=0; i<words; i++)
    state[i] = (((state[i]<<1) | ((state[i] >> (lin_size - 1)) & 1)
               ) & mask);
 }

/* Translate a state in the given direction */

static void translate_direc(unsigned long *state, int direc)
 {
  register int i, j, k;
  unsigned long temp;
  int ofs;

  ofs = 1;
  for(i = 0; i<direc; i++)
    ofs *= lin_size;

  for(i=0; i<ofs; i++)
   for(j=0; j<words; j += (ofs*lin_size))
    {
     temp = state[i+j];
     for(k=0; k<lin_size-1; k++)
       state[i+j+k*ofs] = state[i+j+(k+1)*ofs];
     state[i+j+(lin_size-1)*ofs] = temp;
    }
 }

/* Reflect state along the x-direction */

static void reflect_x(unsigned long *state)
 {
  register int j, x;
  unsigned long tempa, tempb;
  unsigned long mask, rmask;

  for(j=0; j<words; j++)
   {
     tempa = state[j];
     tempb = 0;
     rmask = mask = 1;
     for(x = 0; x<lin_size; x++)
      {
       if(tempa & mask)
         tempb |= rmask;
       mask = (mask << 1);
       if(rmask == 1)
         rmask = 1 << (lin_size-1);
       else
         rmask = (rmask >> 1);
      }
     state[j] = tempb;
   }
 }

/* Reflect state along the y-direction */

static void reflect_y(unsigned long *state)
 {
  register int j;
  unsigned long temp;

  for(j=0; j<words/2; j++)
   {
     temp = state[j];
     state[j] = state[words-j-1];
     state[words-j-1] = temp;
   }
 }

/* Create the momentum basis */

static void create_momenta()
 {
  register long i;
  register int j;
  int flag;
  int direc;
  int *tcnt;
  unsigned long *state;
  int refl;

  if(ddim > 1)
    tcnt = (int *)m_alloc(sizeof(int) * (ddim-1), "create_momenta");

  charge_momentum = (long *)m_alloc(sizeof(long)*nstates, "create_momenta");

  for(i=0; i<nstates; i++)
    charge_momentum[i] = MAGIC;

  pstates = 0;
  for(i=0; i<nstates; i++)
   if(charge_momentum[i] < 0)
    {
     state = index_to_state(i);

     /* Orbit over reflection group */

     for(refl=0; refl<((n_part & 1)?1:4); refl++)
      {

     /* Orbit over d-dimensional translational group */

     for(j=0; j<ddim-1; j++)
       tcnt[j] = 0;
     if(ddim == 1)
      {
       for(j=0; j<lin_size; j += 2)
        {
         charge_momentum[state_to_index(state)] = pstates;
         translate_left(state);
         translate_left(state);
        }
      }
     else
      {
       while(tcnt[ddim-2] < lin_size)
        {
         for(j=0; j<lin_size; j += 2)
          {
           charge_momentum[state_to_index(state)] = pstates;   
           translate_left(state);
           translate_left(state);
          }
         direc = 0;
         translate_direc(state, direc);
         translate_left(state);
         tcnt[direc]++;
         while((tcnt[direc] == lin_size) && (direc < ddim-2))
          {
            tcnt[direc++] = 0;
            translate_direc(state, direc);
            translate_left(state);
            tcnt[direc]++;
          }
        }
      }			/* End of translation group */
       if(refl & 1)
         reflect_y(state);
       else
         reflect_x(state);
      }			/* End of reflection group */
     free(state);
     pstates++;
    }

  if(ddim > 1)
   free(tcnt);

  momentum_charge = (long *)m_alloc(sizeof(long)*pstates, "create_momenta");
  momentum_prefac = (int *)m_alloc(sizeof(int)*pstates, "create_momenta");

  for(i=0; i < pstates; i++)
   {
    momentum_charge[i] = MAGIC;
    momentum_prefac[i] = 0; 
   }

  for(i=0; i<nstates; i++)
   {
     if(momentum_charge[charge_momentum[i]] < 0)
         momentum_charge[charge_momentum[i]] = i;
     momentum_prefac[charge_momentum[i]]++;
   }
 }

/* Apply Hamilton operator */

void apply_internal_ham(struct vector *src, struct vector *dst, char *name)
 {
  register long i, j, dind;
  register int x;
  int nj;
  unsigned long mask, nmask;
  unsigned long tempa, tempb;
  int szsz;
  unsigned long *state;
  REAL prefac;

  if(src->el_cnt != pstates)
   {
    display_error("Dimension of input vector doesn't match - can't do internal matrix application");
    return;
   }

  if(dst->el_cnt != pstates)
   {
    display_error("Dimension of output vector doesn't match - can't do internal matrix application");
    return;
   }

  for(i=0; i<dst->el_cnt; i++)		/* First set destination to zero */
   {
     dst->elements[i].im = dst->elements[i].re = 0.0;
   }

  for(i=0; i<pstates; i++)		/* Loop over all momentum states */
   {
    szsz = 0;
    state = index_to_state(momentum_charge[i]);
    for(j=0; j<words; j++)		/* Loop over all sites */
     {
      mask = 1;
      for(x=0; x<lin_size; x++)
       {
        nmask = mask<<1;		/* x-direction */
        if(x == (lin_size-1))
          nmask = 1;
        tempa = state[j];
        if( ((tempa & mask) && (! (tempa & nmask))) ||
            ((! (tempa & mask)) && (tempa & nmask)) )
         {				/* S^+ S^- - interaction */
           state[j] ^= (mask | nmask);
           dind = charge_momentum[state_to_index(state)];
           prefac = (JVAL/2.0L);
           if(momentum_prefac[dind] != momentum_prefac[i])
              prefac *= sqrt(((REAL) momentum_prefac[i]) /
                             ((REAL) momentum_prefac[dind]));
           dst->elements[dind].re += (prefac * src->elements[i].re);
           dst->elements[dind].im += (prefac * src->elements[i].im);
           state[j] = tempa;
           szsz--;
         }
        else
           szsz++;
        
        if((x + j) & 1)			/* y-direction */
         {
           nj = (j+1) % lin_size;
           tempa = state[j];
           tempb = state[nj];
           if((tempa & mask) != (tempb & mask))
            {				/* S^+ S^- - interaction */
             state[j] ^= mask;
             state[nj] ^= mask;
             dind = charge_momentum[state_to_index(state)];
             prefac = (JVAL/2.0L);
             if(momentum_prefac[dind] != momentum_prefac[i])
                prefac *= sqrt(((REAL) momentum_prefac[i]) /
                               ((REAL) momentum_prefac[dind]));
             dst->elements[dind].re += (prefac * src->elements[i].re);
             dst->elements[dind].im += (prefac * src->elements[i].im);
             state[j] = tempa;
             state[nj] = tempb;
             szsz--;
            }
           else
             szsz++;
         }
        mask = nmask;
       }
     }
    free(state);

    prefac = DELTA/4.0L*((REAL) szsz);		/* The diagonal element */
    dst->elements[i].re += (prefac * src->elements[i].re);
    dst->elements[i].im += (prefac * src->elements[i].im);
   }
 }

/* Init the internal Hamiltonian application */

long init_interal_h()
 {
  register int i;

  fprintf(stderr, "Internal application of the Hamilton-Matrix\n");
  fprintf(stderr, "for a spin-1/2 Heisenberg model on a hexagonal lattice\n");
  fprintf(stderr, "written December 1998/January 1999 by Andreas Honecker\n");

  ddim = 2;			/* This works only in two dimensions */

  words = 1;			/* Compute the number of words needed
				   to represent the system */
  for(i=1; i<ddim; i++)
    words *= lin_size;
  sites = words * lin_size;	/* And the volume of the system */

  init_binomials();

  gen_states();

  fprintf(stderr, "\nNumber of states with %d down spins on a %d^%d lattice: %ld\n",
                   n_part, lin_size, ddim, nstates); 

  create_momenta();

  fprintf(stderr, "\nNumber of associated momentum eigenstates: %ld\n",
                   pstates);

  return(pstates);
 }

/* Exit the internal Hamiltonian application */

void exit_interal_h()
 {
  free_binomials();

  free(momentum_prefac);		/* Now free the tables used */
  free(momentum_charge);
  free(charge_momentum);
  
  free_states();
 }
