/* Standard-Definitions */
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "bessel.h"

/* Test if a real number is zero */

#define ZERO 1e-10

static int is_zero(REAL num)
 {
  if((num < ZERO) && (num > - ZERO))
    return(TRUE);
  return(FALSE);
 }

/* Compute the modified Bessel functions of order 0
   The method is taken from
      M. Abramowitz, I.A. Stegun,
      "Handbook of Mathematical Functions" (1964)
      pages 378/379 ,
   but more accurate coefficients and a few further orders
   are included for improved accuracy */
 
/* Coefficients of
     I_0(x) = \sum_n=0^infinity 1/bessel_I0_coeff(n) x^(2*n)
*/

static REAL bessel_I0_coeffA[] = {
           1.0, 4.0, 64.0, 2304.0, 147456.0, 14745600.0,
           2123366400.0, 416179814400.0, 106542032486400.0,
           34519618525593600.0, 13807847410237440000.0 };

#define NUM_BI0_COEFFA 11

/* Coefficients of
     K_0(x) + ln(x/2) I_0(x) = \sum_n=0^infinity 
            besselK0_coeff(n) x^(2*n)
*/

static REAL bessel_K0_coeffA[] = {
           -0.577215664901532860606512090082,
           0.105696083774616784848371977479,
           0.0144185052359135490530232485925,
           0.000545189960256857844065460609050,
           0.0000102140141359578482579672664609,
           0.115703509415134038135228220165e-6,
           0.881988306445118063181883216161e-9,
           0.484319856036633860475945527447e-11,
           0.200919902502222403532555803418e-13,
           0.652310971338580262711289055153e-16,
           0.170320001314837844556920771047e-18 };

#define NUM_BK0_COEFFA 11

/* The following coefficients are precisely from Abramowitz and Stegun */

/* 9.8.2 */
static REAL bessel_I0_coeffB[] = {
               0.39894228, 0.01328592, 0.00225319, -0.00157565,
               0.00916281, -0.02057706, 0.02635537, -0.01647633,
               0.00392377 };

#define NUM_BI0_COEFFB 9

/* 9.8.6 */

static REAL bessel_K0_coeffB[] = {
               1.25331414, -0.07832358, 0.02189568, -0.01062446,
               0.00587872, -0.00251540, 0.00053208 };

#define NUM_BK0_COEFFB 7

/* The Bessel-I_0-function of a real argument */

REAL evaluate_bessel_I0(REAL x)
 {
  REAL sum, xsq, tinv;
  int i;

  if(x < 0)		/* I_0(-x) = I_0(x) */
    x = -x;		/* So, consider only positive arguments */

  if(x <= 15.0/4.0)
   {
    xsq = x*x;
    sum = 0;
    for(i=NUM_BI0_COEFFA-1; i>=0; i--)
       sum = 1.0/bessel_I0_coeffA[i] + sum*xsq;
    return(sum); 
   }
  else
   {
    tinv = 15.0/4.0/x;
    sum = 0;
    for(i=NUM_BI0_COEFFB-1; i>=0; i--)
       sum = bessel_I0_coeffB[i] + sum*tinv;
    return(sum/sqrt(x)*exp(x));
   }
 }

/* The Bessel-K_0-function of a positive real argument */

REAL evaluate_bessel_K0(REAL x)
 {
  REAL sum, xsq, tinv;
  int i;

  if(x <= 2.0)
   {
    xsq = x*x;
    sum = 0;
    for(i=NUM_BK0_COEFFA-1; i>=0; i--)
       sum = bessel_K0_coeffA[i] + sum*xsq;
    return(-log(x/2.0)*evaluate_bessel_I0(x)+sum);
   }
  else
   {
    tinv = 2.0/x;
    sum = 0;
    for(i=NUM_BK0_COEFFB-1; i>=0; i--)
       sum = bessel_K0_coeffB[i] + sum*tinv;
    return(sum/sqrt(x)*exp(-x));
   }
 }

/* The Bessel-I_0-function with a complex argument */

void bessel_I0(COMPLEX *x, COMPLEX *result)
 {
  if(! is_zero(x->im))
    display_error("I_0(x) defined only for real x");

  result->re = evaluate_bessel_I0(x->re);
  result->im = 0;
 }

/* The Bessel-K_0-function with a complex argument */

void bessel_K0(COMPLEX *x, COMPLEX *result)
 {
  if(! is_zero(x->im))
    display_error("K_0(x) defined only for real x");
  if(x->re <= 0)
    display_error("K_0(x) defined only for x>0");

  result->re = evaluate_bessel_K0(x->re);
  result->im = 0;
 }

