#include <stdlib.h>
#include <stdio.h>
#include "useful.h"
#include "short_alloc.h"
#include "complex.h"

/* Special handling of allocation of "short" memory segments */

/* Definitions */

#define INTBITS (8*sizeof(unsigned int))	/* Number of bits in unsigned int data type */
#define NFLAGS 512				/* Size of flag array per block */
#define	NELBLK (NFLAGS*INTBITS)			/* Number of elements per block */

#define FULL 0xffffffff				/* 2^INTBITS - 1 */

/* Structure of a block descriptor */

struct block {struct block *next;
              BYTE *data;
              unsigned int flags[NFLAGS];
              int nalloc_el;
             };

/* List of blocks */

static struct block *blocks[MAXSHORT+1];

#ifdef DEBUGALLOCDETAIL
 static long alloc_calls[MAXSHORT+1];
#endif

/* Initialize this allocator */

void init_short_malloc()
 {
  register int i;

  for(i=0; i<=MAXSHORT; i++)
   {
    blocks[i]=0;
#ifdef DEBUGALLOCDETAIL
    alloc_calls[i]=0;
#endif
   }
 }

/* *** Allocation *** */

/* Allocate an element in a block */

static BYTE *alloc_in_block(struct block *ptr, int size)
 {
  register int i, j;
  unsigned int mask;

  for(i=0; i<NFLAGS; i++)
    if(ptr->flags[i] != FULL)
      {
        mask = 1;
        for(j=0; j<INTBITS; j++)
         {
          if(! (ptr->flags[i] & mask))
           {
             ptr->nalloc_el++;
             ptr->flags[i] |= mask;
#ifdef DEBUGALLOCDETAIL
 printf("Occupying element %d - mask is now %x\n", (i*INTBITS)+j, ptr->flags[i]);
#endif
             return(ptr->data + size*((i*INTBITS)+j));
           }
          mask = (mask << 1);
         }
        display_error("Internal error: No free bit found in alloc_in_block()");
        display_error("Check proper definition of sizes and \"FULL\" !");
        return(0); 
      }
  display_error("Internal error: No free flag found in alloc_in_block()");

#ifdef DEBUGALLOCDETAIL
  printf("%d\n", ptr->nalloc_el);
  for(i=0; i<NFLAGS; i++)
    printf("%d: %x\n", i, ptr->flags[i]);
#endif

  exit(17);
 }

/* Create a new block */

static int create_new_block(int size)
 {
  struct block *tptr;
  register int i;

#ifdef DEBUGALLOC
  printf("Creating new block of size %d\n", size);
#endif

  tptr = (struct block *)malloc(sizeof(struct block));
  if(! tptr)
    return(FALSE);

  tptr->data = (BYTE *)malign16(size*NELBLK);
  if(! tptr->data)
    {
      free(tptr);
      return(FALSE);
    }

  for(i=0; i<NFLAGS; i++)
    tptr->flags[i] = 0;

  tptr->nalloc_el = 0;
  tptr->next = blocks[size];
  blocks[size] = tptr;

  return(TRUE);
 }

/* The main allocation function */

BYTE *short_malloc(int size)
 {
  struct block *ptr;

#ifdef SSE
  /* Make it a multiple of 16 */
  size = ((size + 0xf) & 0xfff0 );
#endif

#ifdef DEBUGALLOC
  printf("Call: short_malloc(%d)\n", size);
#endif
#ifdef DEBUGALLOCDETAIL
  alloc_calls[size]++;
  printf("\t(%ld calls of this size so far)\n", alloc_calls[size]);
#endif

  for(ptr=blocks[size]; ptr; ptr=ptr->next)
    if(ptr->nalloc_el < NELBLK)
      return(alloc_in_block(ptr, size));

  if(! create_new_block(size))
     return(0);
#ifdef DEBUGALLOCDETAIL
  alloc_calls[size]--;
#endif
  return(short_malloc(size)); 
 }

/* *** Freeing *** */

/* Free memory in a specific block */

static void free_in_block(struct block *tptr, BYTE *ptr, int size)
 {
  int index;
  int i, j;
  unsigned int mask;
  struct block *hptr;

#ifdef DEBUGFREE
  printf("Call: free_in_block(%lx, %lx, %d)\n", tptr, ptr, size);
#endif

  index = ((int) (ptr - tptr->data))/size; 
  i = index / INTBITS;
  j = index % INTBITS;

  mask = 1<<j;
  if(! (tptr->flags[i] & mask))
   {
    display_error("Internal error: Space not marked as occupied in free_in_block()");
    return;
   }

  tptr->flags[i] ^= mask;
  tptr->nalloc_el--;
#ifdef DEBUGFREE
  printf("%d elements now allocated in this block\n", tptr->nalloc_el);
#endif

  if(tptr->nalloc_el)
    return;

  /* Free the block */
#ifdef DEBUGFREE
  printf("freeing this block\n");
#endif
  free(tptr->data);

  if(tptr == blocks[size])
    blocks[size] = tptr->next;
  else
    for(hptr=blocks[size]; hptr; hptr=hptr->next)
      if(hptr->next == tptr)
        hptr->next = tptr->next;

  free(tptr); 
 }

/* Free allocated memory */

void short_free(BYTE *ptr, int size)
 {
  struct block *tptr;
  char tmp_str[64];

#ifdef DEBUGFREE
  printf("Call: short_free(%lx, %d)\n", ptr, size);
#endif

  for(tptr=blocks[size]; tptr; tptr=tptr->next)
   {
    if(tptr->data <= ptr)
     if((tptr->data + size*NELBLK) > ptr)
      {
       free_in_block(tptr, ptr, size);
       return;
      }
   }
  display_error("Internal error: Memory segment to free not identified in short_free()");
  sprintf(tmp_str, "ptr = %lx, size=%d", ptr, size);
  display_error(tmp_str);
  fprintf(stderr, "ptr = %lx, size=%d\n", ptr, size);
 }

/* Free all allocated memory */

void short_freeall()
 {
  struct block *tptr, *hptr;
  register int i;

  for(i=0; i<=MAXSHORT; i++)
   {
    for(tptr=blocks[i]; tptr; tptr=hptr)
     {
      hptr = tptr->next;
      free(tptr->data);
      free(tptr);
     }
   }

  init_short_malloc();
 }
